"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from . import utils
import numpy as np
from . import models
import sys
import os
import mrcfile
import tifffile
import pathlib

def predict(
    input_path,
    output_path,
    model_path,
    model="unet",
    patch_size=(1024, 1024),
    padding=15,
    batch_size=4,
    output_resize_to=None
):
    if model == "unet":
        model = models.get_model_unet(input_size=patch_size)
        model.load_weights(model_path)
    else:
        print("Not supported model", model, "Stop")
        sys.exit(0)

    predict_dir(
        input_path=input_path,
        output_path=output_path,
        model=model,
        patch_size=patch_size,
        padding=padding,
        batch_size=batch_size,
        output_resize_to=output_resize_to,
    )

def predict_dir(
    input_path,
    output_path,
    model,
    patch_size=(1024, 1024),
    padding=15,
    batch_size=4,
    output_resize_to=None
):
    """
    Denoises images / movies
    :param input_path: Input path to directory with movies / averages to denoise.
    :param output_path: Folder where results should be written.
    :param model: Trained model
    :param patch_size: Patch size in Pixel. Image will be denoised in patches and then stitched together.
    :param padding: Padding to remove edge effects.
    :param batch_size: Number of patches denoised in parallel.
    :param output_resize_to: The denoised image will be downsized to this dimension.
    :return: List of paths to denoised images
    """
    list_files = []
    for (dirpath, dirnames, filenames) in os.walk(input_path):
        for filename in filenames:
            if filename.endswith(utils.SUPPORTED_FILES):
                path = os.path.join(dirpath, filename)
                list_files.append(path)

    denoise_image_paths = predict_list(
        list_files,
        output_path,
        model,
        patch_size=patch_size,
        padding=padding,
        batch_size=batch_size,
        output_resize_to=output_resize_to
    )

    return denoise_image_paths

def predict_list(
    image_paths,
    output_path,
    model,
    patch_size=(1024, 1024),
    padding=15,
    batch_size=4,
    output_resize_to=None,
    squarify=False,
):
    """
    Denoises images / movies
    :param image_paths: List of images to filter
    :param output_path: Folder where results should be written.
    :param model: Trained model
    :param patch_size: Patch size in Pixel. Image will be denoised in patches and then stitched together.
    :param padding: Padding to remove edge effects.
    :param batch_size: Number of patches denoised in parallel.
    :param output_resize_to: The denoised image will be downsized to this dimension. In case of a
    list it should be [height,width]. If a single number if passed the short image side is scaled to
    this size and the long side according the aspect ratio.
    :return: List of paths to denoised images
    """

    denoise_image_paths = []
    for path in image_paths:
        if path.endswith(utils.SUPPORTED_FILES):
            try:
                os.makedirs(output_path)
            except FileExistsError:
                pass
            pos_path = pathlib.PurePath(path)
            file_outpath = os.path.join(output_path, pos_path.parent.name)
            if not os.path.exists(file_outpath):
                os.makedirs(file_outpath)

            opath = os.path.join(file_outpath, pos_path.name)

            if not os.path.exists(opath):
                if utils.is_movie(path):
                    even, odd = utils.create_image_pair(path)
                    denoised_even = predict_np(
                        model,
                        even,
                        patch_size=patch_size,
                        padding=padding,
                        batch_size=batch_size,
                    )
                    denoised_odd = predict_np(
                        model,
                        odd,
                        patch_size=patch_size,
                        padding=padding,
                        batch_size=batch_size,
                    )
                    denoised = denoised_even + denoised_odd
                else:
                    img = utils.read_image(path)
                    img = img.squeeze()
                    denoised = predict_np(
                        model,
                        image=img,
                        patch_size=patch_size,
                        padding=padding,
                        batch_size=batch_size,
                    )
                # Write result to disk

                if output_resize_to is not None:
                    if squarify:
                        denoised = squarify(denoised)
                    resize_to = output_resize_to
                    if not isinstance(output_resize_to, (list, tuple)):
                        # Rescale shorter side to output_resize_to
                        ar = denoised.shape[0] / denoised.shape[1]
                        if ar < 1:
                            height = output_resize_to
                            width = int(output_resize_to / ar)
                        else:
                            height = int(output_resize_to * ar)
                            width = output_resize_to
                        resize_to = [height, width]

                    from PIL import Image
                    denoised = np.array(Image.fromarray(denoised).resize(
                        (resize_to[1], resize_to[0]), resample=Image.BILINEAR))

                print("Write denoised image in", opath)
                if opath.endswith((".mrc", ".mrcs")):
                    with mrcfile.new(opath, overwrite=True) as mrc:
                        mrc.set_data(denoised)
                elif opath.endswith((".tif", ".tiff")):
                    tifffile.imwrite(opath, denoised)
            else:
                print("Already filtered:", path, "Skip!")
            denoise_image_paths.append(opath)

    return denoise_image_paths

def squarify(image, size=None):
    np.random.seed()
    if size is not None:
        target_size=size
    else:
        target_size = np.max(image.shape)
    mean = np.mean(image)
    rectified_image = np.ones(shape=(target_size, target_size)) * mean
    rectified_image[(rectified_image.shape[0]-image.shape[0]):rectified_image.shape[0],0:image.shape[1]] = image

    return rectified_image

def predict_np(model, image, patch_size=(1024, 1024), padding=15, batch_size=4):
    """
    Denoises an image given a keras model.
    :param model: Trained noise2noise model
    :param image: Image as 2D numpy array
    :param patch_size: Patch size in Pixel. Image will be denoised in patches and then stitched together.
    :param padding: Padding to remove edge effects.
    :param batch_size: Number of patches denoised in parallel.
    :return: Denoised image (2D numpy array)
    """
    if image.ndim != 2:
        print("Your image should have only two dimensions. Return none")
        return None
    image, mean, sd = utils.normalize(image)
    img_patches, pads = utils.image_to_patches(
        image, patch_size=patch_size, padding=padding
    )
    denoised_patches = model.predict(
        x=img_patches[:, :, :, np.newaxis], batch_size=batch_size
    )

    denoised_micrograph = utils.patches_to_image(
        denoised_patches, pads, image_shape=image.shape, padding=padding
    )
    denoised_micrograph = denoised_micrograph.astype(np.float32)
    denoised_micrograph = denoised_micrograph * sd + mean
    return denoised_micrograph



"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import sys
import json
import os
import h5py
from gooey import Gooey, GooeyParser
import janni.__init__ as ini

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
try:
    os.environ["CUDA_VISIBLE_DEVICES"]
except KeyError:
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

DEFAULT_BATCH_SIZE = 4
DEFAULT_PADDING = 24

ARGPARSER = None

def create_config_parser(parser):
    config_required_group = parser.add_argument_group(
        "Required arguments",
        "The arguments are required to create a config file for JANNI",
    )

    config_required_group.add_argument(
        "config_out_path",
        default="config_janni.json",
        help="Path where you want to write the config file.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("json")',
                "message": "File has to end with .json!",
            },
            "default_file": "config_janni.json"
        },
    )

    config_required_group.add_argument(
        "--movie_dir",
        help="Path to the directory with the movie files. The movie files can be unaligned. You should use at least 30 movies. If an average with the same filename already exists in even_dir or odd_dir it will be skipped.",
        widget="DirChooser",
    )

    config_required_group.add_argument(
        "--even_dir",
        help="For each movie in movie_dir, an average based on the even frames is calculated and saved in even_dir.",
        widget="DirChooser",
    )

    config_required_group.add_argument(
        "--odd_dir",
        help="For each movie in movie_dir, an average based on the odd frames is calculated and saved in odd_dir.",
        widget="DirChooser",
    )

    config_required_group.add_argument(
        "--saved_weights_name",
        default="janni_model.h5",
        help="Path for saving final weights.",
        widget="FileSaver",
        gooey_options={
            "validator": {
                "test": 'user_input.endswith("h5")',
                "message": "File has to end with .h5!",
            },
            "default_file": "janni_model.h5"
        },
    )


    config_optional_group = parser.add_argument_group(
        "Optional arguments",
        "The arguments are optional to create a config file for JANNI",
    )

    config_optional_group.add_argument(
        "--patch_size",
        default=1024,
        type=int,
        help="The image will be denoised in patches. This field defines the patch size..",
    )

    config_optional_group.add_argument(
        "--batch_size",
        type=int,
        default=4,
        help="How many patches are in one mini-batch. If you have memory problems (e.g with cards < 8GB memory), you can try to reduce this value.",
    )

    config_optional_group.add_argument(
        "--learning_rate",
        type=float,
        default=10**-3,
        help="Learning rate, should not be changed.",
    )

    config_optional_group.add_argument(
        "--nb_epoch",
        type=int,
        default=100,
        help="Number of epochs to train. Default is 100. More epochs seems to only slightly improve the results.",
    )


def create_train_parser(parser):
    required_group = parser.add_argument_group(
        "Required arguments", "These options are mandatory to train JANNI"
    )

    required_group.add_argument(
        "config_path",
        help="Path to config.json",
        widget="FileChooser",
        gooey_options={
            "wildcard": "*.json"
        }
    )

    optional_group = parser.add_argument_group(
        "Optional arguments", "These options are optional to train JANNI"
    )

    optional_group.add_argument(
        "-g", "--gpu", type=int, default=-1, help="GPU ID to run on"
    )

def create_predict_parser(parser):
    required_group = parser.add_argument_group(
        "Required arguments", "These options are mandatory to run JANNI"
    )

    required_group.add_argument(
        "input_path",
        help="Directory / file path with images/movies to denoise. In our experience movie aligned averages are working best. \n",
        widget="DirChooser",
    )

    required_group.add_argument(
        "output_path",
        help="Directory / file path to write denoised images.\n",
        widget="DirChooser",
    )
    required_group.add_argument(
        "model_path",
        help="File path to trained model.",
        widget="FileChooser",
        gooey_options={
            "wildcard": "*.h5"
        }
    )

    optional_group = parser.add_argument_group(
        "Optional arguments", "These options are optional to run JANNI"
    )
    optional_group.add_argument(
        "-ol",
        "--overlap",
        help="The patches have to overlap to remove artifacts. This is the amount of overlap in pixel. If you observe a grid like pattern in your images, increase this value.",
        default=DEFAULT_PADDING,
    )
    optional_group.add_argument(
        "-bs",
        "--batch_size",
        help="Number of patches predicted in parallel\n",
        default=DEFAULT_BATCH_SIZE,
    )
    optional_group.add_argument(
        "-g", "--gpu", type=int, default=-1, help="GPU ID to run on"
    )

def create_parser(parser):

    subparsers = parser.add_subparsers(help="sub-command help")

    parser_config= subparsers.add_parser("config", help="Create the configuration file for JANNI")
    create_config_parser(parser_config)

    parser_train = subparsers.add_parser("train", help="Train JANNI for your dataset.")
    create_train_parser(parser_train)

    parser_predict = subparsers.add_parser("denoise", help="Denoise micrographs using a (pre)trained model.")
    create_predict_parser(parser_predict)



def get_parser():
    parser = GooeyParser(description="Just another noise to noise implementation")
    create_parser(parser)
    return parser


def _main_():
    global ARGPARSER
    import sys

    if len(sys.argv) >= 2:
        if not "--ignore-gooey" in sys.argv:
            sys.argv.append("--ignore-gooey")

    kwargs = {"terminal_font_family": "monospace", "richtext_controls": True}
    Gooey(
        main,
        program_name="JANNI " + ini.__version__,
        #image_dir=os.path.join(os.path.abspath(os.path.dirname(__file__)), "../icons"),
        progress_regex=r"^.* \( Progress:\s+(-?\d+) % \)$",
        disable_progress_bar_animation=True,
        tabbed_groups=True,
        default_size=(1024, 730),
        **kwargs
    )()


def main(args=None):

    if args is None:
        parser = get_parser()
        args = parser.parse_args()



    if "config" in sys.argv[1]:
        generate_config_file(config_out_path=args.config_out_path,
                             architecture="unet",
                             patch_size=args.patch_size,
                             movie_dir=args.movie_dir,
                             even_dir=args.even_dir,
                             odd_dir=args.odd_dir,
                             batch_size=args.batch_size,
                             learning_rate=args.learning_rate,
                             nb_epoch=args.nb_epoch,
                             saved_weights_name=args.saved_weights_name)
    else:
        if isinstance(args.gpu, list):
            if len(args.gpu) == 1:
                if args.gpu[0] != "-1":
                    str_gpus = args.gpu[0].strip().split(" ")
                    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
        elif args.gpu != -1:
            str_gpus = str(args.gpu)
            os.environ["CUDA_VISIBLE_DEVICES"] = str_gpus

        if "train" in sys.argv[1]:
            config = read_config(args.config_path)

            from . import train

            train.train(
                even_path=config["train"]["even_dir"],
                odd_path=config["train"]["odd_dir"],
                model_out_path=config["train"]["saved_weights_name"],
                movie_path=config["train"]["movie_dir"],
                learning_rate=config["train"]["learning_rate"],
                epochs=config["train"]["nb_epoch"],
                model=config["model"]["architecture"],
                patch_size=(config["model"]["patch_size"], config["model"]["patch_size"]),
                batch_size=config["train"]["batch_size"],
            )

        elif "denoise" in sys.argv[1]:

            input_path = args.input_path
            output_path = args.output_path
            model_path = args.model_path
            from . import predict

            batch_size = DEFAULT_BATCH_SIZE
            padding = DEFAULT_PADDING

            with h5py.File(model_path, mode="r") as f:
                try:
                    import numpy as np

                    model = str(np.array((f["model_name"])))
                    patch_size = tuple(f["patch_size"])
                except KeyError:
                    print("Error on loading model", model_path)
                    sys.exit(0)

            if args.overlap is not None:
                padding = int(args.overlap)

            if args.batch_size is not None:
                batch_size = int(args.batch_size)

            predict.predict(
                input_path=input_path,
                output_path=output_path,
                model_path=model_path,
                model=model,
                patch_size=patch_size,
                padding=padding,
                batch_size=batch_size,
            )

def generate_config_file(config_out_path,
                         architecture,
                         patch_size,
                         movie_dir,
                         even_dir,
                         odd_dir,
                         batch_size,
                         learning_rate,
                         nb_epoch,
                         saved_weights_name):
    model_dict = {'architecture': architecture,
                  'patch_size': patch_size,
                  }

    train_dict = {'movie_dir': movie_dir,
                  'even_dir': even_dir,
                  'odd_dir': odd_dir,
                  'batch_size': batch_size,
                  'learning_rate': learning_rate,
                  'nb_epoch': nb_epoch,
                  "saved_weights_name": saved_weights_name,
                  }

    from json import dump
    dict = {"model": model_dict, "train": train_dict}
    with open(config_out_path, 'w') as f:
        dump(dict, f, ensure_ascii=False, indent=4)
    print("Wrote config to", config_out_path)

def read_config(config_path):
    with open(config_path) as config_buffer:
        try:
            config = json.loads(config_buffer.read())
        except json.JSONDecodeError:
            print(
                "Your configuration file seems to be corruped. Please check if it is valid."
            )
    return config


if __name__ == "__main__":
    _main_()



"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from keras.models import Model
from keras.layers import Input, Add, Conv2DTranspose, MaxPooling2D, UpSampling2D, ReLU
from keras.layers.convolutional import Conv2D
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.merge import concatenate


def get_rednet(
    filter_size=(3, 3),
    num_filters=64,
    num_conv_layers=15,
    input_size=(128, 128),
    use_skip=True,
):
    inputs = Input(shape=(input_size[0], input_size[1], 1))

    # Encoder
    skip_conns = []
    x = Conv2D(
        num_filters, filter_size, padding="same", kernel_initializer="he_normal"
    )(inputs)
    x = ReLU()(x)
    for k in range(1, num_conv_layers):
        x = Conv2D(
            num_filters, filter_size, padding="same", kernel_initializer="he_normal"
        )(x)
        x = ReLU()(x)
        if (k + 1) % 2 == 0 and use_skip:
            skip_conns.append(x)

    # Decoder
    for k in range(num_conv_layers - 1):
        x = Conv2DTranspose(
            num_filters, filter_size, padding="same", kernel_initializer="he_normal"
        )(x)
        x = ReLU()(x)
        if (k + 1) % 2 == 0 and use_skip:
            # if k != 13 and k != 11:
            skip = skip_conns.pop()
            x = Add()([x, skip])
            x = ReLU()(x)

    x = Conv2DTranspose(1, filter_size, padding="same", kernel_initializer="he_normal")(
        x
    )
    model = Model(inputs=inputs, outputs=x)
    return model


def get_model_unet(input_size=(1024, 1024), kernel_size=(3, 3)):
    inputs = Input(shape=(input_size[0], input_size[1], 1))
    skips = [inputs]

    x = Conv2D(
        name="enc_conv0",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(inputs)
    x = LeakyReLU(alpha=0.1)(x)
    x = Conv2D(
        name="enc_conv1",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = MaxPooling2D((2, 2))(x)  # --- pool_1
    skips.append(x)

    x = Conv2D(
        name="enc_conv2",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = MaxPooling2D((2, 2))(x)  # --- pool_2
    skips.append(x)

    x = Conv2D(
        name="enc_conv3",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = MaxPooling2D((2, 2))(x)  # --- pool_3
    skips.append(x)

    x = Conv2D(
        name="enc_conv4",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = MaxPooling2D((2, 2))(x)  # --- pool_4
    skips.append(x)

    x = Conv2D(
        name="enc_conv5",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = MaxPooling2D((2, 2))(x)  # --- pool_5 (not re-used)

    x = Conv2D(
        name="enc_conv6",
        filters=48,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)

    x = UpSampling2D((2, 2))(x)

    x = concatenate([x, skips.pop()])
    x = Conv2D(
        name="dec_conv5",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)

    x = Conv2D(
        name="dec_conv5b",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = UpSampling2D((2, 2))(x)

    x = concatenate([x, skips.pop()])

    x = Conv2D(
        name="dec_conv4",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = Conv2D(
        name="dec_conv4b",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)

    x = UpSampling2D((2, 2))(x)

    x = concatenate([x, skips.pop()])

    x = Conv2D(
        name="dec_conv3",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = Conv2D(
        name="dec_conv3b",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = UpSampling2D((2, 2))(x)

    x = concatenate([x, skips.pop()])
    x = Conv2D(
        name="dec_conv2",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = Conv2D(
        name="dec_conv2b",
        filters=96,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)

    x = UpSampling2D((2, 2))(x)
    x = concatenate([x, skips.pop()])

    x = Conv2D(
        name="dec_conv1a",
        filters=64,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = Conv2D(
        name="dec_conv1b",
        filters=32,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    x = LeakyReLU(alpha=0.1)(x)

    outputs = Conv2D(
        filters=1,
        kernel_size=kernel_size,
        padding="same",
        kernel_initializer="he_normal",
    )(x)
    model = Model(inputs=inputs, outputs=outputs)
    return model



"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from . import models
from . import patch_pair_generator as gen
from keras.optimizers import Adam
import os
from . import utils
import mrcfile
import tifffile
import numpy as np

def train(
    even_path,
    odd_path,
    model_out_path,
    movie_path=None,
    learning_rate=0.001,
    epochs=50,
    model="unet",
    patch_size=(1024, 1024),
    batch_size=4,
):
    '''
    Does the complete noise2noise training and writes the model file to disk.
    :param even_path: Path where "even averages" will be written.
    :param odd_path: Path here "odd averages" will be written
    :param model_out_path: Filepath where model will be written.
    :param movie_path: Path to movie files. Supported are .mrc, .mrcs, .tiff and .tif.
    :param learning_rate: Learning rate used during training.
    :param epochs: Number of epochs to train the network
    :param model: Model indentifier. Right now only "unet" is supported.
    :param patch_size: Patch size in pixel. The network is trained on random patches of the images.
    :param batch_size: Mini-batch size used during training.
    :return: trained model
    '''

    print("Start training")
    # Read training even/odd micrographs
    trained_model = train_movie_dir(
        even_path=even_path,
        odd_path=odd_path,
        movie_path=movie_path,
        learning_rate=learning_rate,
        epochs=epochs,
        model=model,
        patch_size=patch_size,
        batch_size=batch_size,
        )
    trained_model.save_weights(model_out_path)
    import h5py
    with h5py.File(model_out_path, mode='r+') as f:
        f["model_name"] = np.array((model), dtype=h5py.special_dtype(vlen=str))
        f["patch_size"] = patch_size

    print("Training done. Weights saved to " + model_out_path)

    return trained_model

def train_movie_dir(
    even_path,
    odd_path,
    movie_path=None,
    learning_rate=0.001,
    epochs=50,
    model="unet",
    patch_size=(1024, 1024),
    batch_size=4,
):
    '''
    Does the complete noise2noise training.
    :param even_path: Path where "even averages" will be written.
    :param odd_path: Path here "odd averages" will be written
    :param model_out_path: Filepath where model will be written.
    :param movie_path: Path to movie files. Supported are .mrc, .mrcs, .tiff and .tif.
    :param learning_rate: Learning rate used during training.
    :param epochs: Number of epochs to train the network
    :param model: Model indentifier. Right now only "unet" is supported.
    :param patch_size: Patch size in pixel. The network is trained on random patches of the images.
    :param batch_size: n
    :return: trained model
    '''

    # Read training even/odd micrographs
    even_files, odd_files = calc_even_odd(
        movie_path, even_path, odd_path, recursive=True
    )

    trained_model = train_pairs(
        even_files,
        odd_files,
        model=model,
        learning_rate=learning_rate,
        patch_size=patch_size,
        batch_size=batch_size,
        epochs=epochs,
        valid_split=0.1,
    )
    return trained_model



def calc_even_odd(movie_path, even_path, odd_path, recursive=True):
    """
    Calculates averages based on the even/odd frames of the movies in movie_path and save the
    respective averages in even_path or odd_path.
    :param movie_path: Path to movie files. Supported are .mrc, .mrcs, .tiff and .tif.
    :param even_path: Path where "even averages" will be written.
    :param odd_path: Path here "odd averages" will be written
    :param recursive: If true, the movie_path is scanned recurively for movies.
    :return: Two lists with even filepaths and odd filepaths.
    """

    # Read training even/odd micrographs
    even_files = []
    odd_files = []
    for (dirpath, dirnames, filenames) in os.walk(even_path):
        for filename in filenames:
            if filename.endswith(utils.SUPPORTED_FILES):
                even_files.append(os.path.join(dirpath, filename))

    for (dirpath, dirnames, filenames) in os.walk(odd_path):
        for filename in filenames:
            if filename.endswith(utils.SUPPORTED_FILES):
                odd_files.append(os.path.join(dirpath, filename))

    # Create training data for movies which are not in even/odd directory
    try:
        os.makedirs(even_path)
    except FileExistsError:
        pass
    try:
        os.makedirs(odd_path)
    except FileExistsError:
        pass

    movies_to_split = []
    filenames_even = list(map(os.path.basename, even_files))
    filenames_odd = list(map(os.path.basename, odd_files))
    if movie_path:
        for (dirpath, dirnames, filenames) in os.walk(movie_path):
            for filename in filenames:
                if filename.endswith(utils.SUPPORTED_FILES):
                    if filename not in filenames_even and filename not in filenames_odd:
                        path = os.path.join(dirpath, filename)
                        movies_to_split.append((path,filename))

            if recursive == False:
                break

    for tuble_index, movie_tuble in enumerate(movies_to_split):
        path, filename = movie_tuble
        print("Create even/odd average for:", path, "( Progress: ",int(100*tuble_index/len(movies_to_split)),"% )")
        even, odd = utils.create_image_pair(path)
        out_even_path = os.path.join(even_path, filename)
        out_odd_path = os.path.join(odd_path, filename)
        if path.endswith(("mrcs", "mrc")):
            with mrcfile.new(out_even_path, overwrite=True) as mrc:
                mrc.set_data(even)

            with mrcfile.new(out_odd_path, overwrite=True) as mrc:
                mrc.set_data(odd)

        elif path.endswith((".tif", ".tiff")):
            tifffile.imwrite(out_even_path, even)
            tifffile.imwrite(out_odd_path, odd)

        even_files.append(out_even_path)
        odd_files.append(out_odd_path)
        filenames_even.append(filename)
        filenames_odd.append(filename)

    print("Reset progress bar: ( Progress: -1 % )")
    return even_files, odd_files


def train_pairs(
    pair_files_a,
    pari_files_b,
    model="unet",
    learning_rate=0.001,
    epochs=50,
    patch_size=(1024, 1024),
    callbacks=[],
    batch_size=4,
    valid_split=0.1,
    loss="mse"
):
    """
    Training noise2noise model.
    :param pair_files_a: List with paths to the first images of the image pairs.
    :param pari_files_b: List with paths to the second images of the image pairs.
    :param model: Model indentifier. Right now only "unet" is supported.
    :param learning_rate: Learning rate used during training.
    :param epochs: Number of epochs to train the network
    :param patch_size: Patch size in pixel. The network is trained on random patches of the images.
    :param callbacks: Optional callbacks during training. See Keras callbacks for more information.
    :param batch_size: Mini-batch size used during training
    :param valid_split: training-validion split.
    :return: Trained keras model
    """
    train_valid_split = int(valid_split * len(pair_files_a))
    train_pair_a_files = pair_files_a[train_valid_split:]
    valid_pair_a_files = pair_files_a[:train_valid_split]
    train_pair_b_files = pari_files_b[train_valid_split:]
    valid_pair_b_files = pari_files_b[:train_valid_split]

    train_gen = gen.patch_pair_batch_generator(
        pair_a_images=train_pair_a_files,
        pair_b_images=train_pair_b_files,
        patch_size=patch_size,
        batch_size=batch_size,
        augment=True,
    )

    valid_gen = gen.patch_pair_batch_generator(
        pair_a_images=valid_pair_a_files,
        pair_b_images=valid_pair_b_files,
        patch_size=patch_size,
        batch_size=batch_size,
    )

    if model == "unet":
        model = models.get_model_unet(input_size=patch_size, kernel_size=(3, 3))
    opt = Adam(lr=learning_rate, epsilon=10 ** -8, amsgrad=True)
    model.compile(optimizer=opt, loss=loss)

    history = model.fit_generator(
        generator=train_gen,
        validation_data=valid_gen,
        epochs=epochs,
        callbacks=callbacks,
        workers=4,
        use_multiprocessing=True
    )
    return model



"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import mrcfile
import tifffile
import numpy as np
from . import utils

SUPPORTED_FILES = (".mrc", ".mrcs", ".tiff", ".tif")


def image_to_patches(image, patch_size=(1024, 1024), padding=15):
    '''
    Divides an image into patches
    :param image: 2D numpy array
    :param patch_size: Size of patches in pixel
    :param padding: Number of pixel the patches do overlap.
    :return: 3D Numpy array with shape (NUM_PATCHES,PATCH_WIDTH,PATCH_HIGHT) and applied pads
    '''
    roi_size = (patch_size[0] - 2 * padding, patch_size[1] - 2 * padding)

    pad_before0 = padding
    diff0 = image.shape[0] - roi_size[0] * (image.shape[0] // roi_size[0])
    if diff0 == 0:
        pad_after0 = 0
    else:
        pad_after0 = patch_size[0] - diff0

    pad_before1 = padding
    diff1 = image.shape[1] - roi_size[1] * (image.shape[1] // roi_size[1])
    if diff1 == 0:
        pad_after1 = 0
    else:
        pad_after1 = patch_size[1] - (
            image.shape[1] - roi_size[1] * (image.shape[1] // roi_size[1])
        )
    pads = [(pad_before0, pad_after0), (pad_before1, pad_after1)]

    n0 = int(np.ceil(image.shape[0] / roi_size[0]))
    n1 = int(np.ceil(image.shape[1] / roi_size[0]))

    image = np.pad(image, pads, mode="symmetric")

    total = int(n0 * n1)
    patches = np.zeros(shape=(total, patch_size[0], patch_size[1]), dtype=np.float32)

    entry_index = 0
    for off0 in range(n0):
        for off1 in range(n1):
            start0 = off0 * roi_size[0]
            end0 = start0 + patch_size[0]
            start1 = off1 * roi_size[1]
            end1 = start1 + patch_size[1]
            patches[entry_index] = image[start0:end0, start1:end1]
            entry_index = entry_index + 1

    return patches, pads


def patches_to_image(patches, pads, image_shape=(4096, 4096), padding=15):
    '''
    Stitches the image together given the patches.
    :param patches: 3D numpy array with shape (NUM_PATCHES,PATCH_WIDTH,PATCH_HIGHT)
    :param pads: Applied pads
    :param image_shape: Original image size
    :param padding: Specified padding
    :return: Image as 2D numpy array
    '''
    patch_size = (patches.shape[1], patches.shape[2])

    roi_size = (patch_size[0] - 2 * padding, patch_size[1] - 2 * padding)

    entry_index = 0
    image = np.zeros(
        shape=(
            image_shape[0] + pads[0][0] + pads[0][1],
            image_shape[1] + pads[1][0] + pads[1][1],
        )
    )
    n0 = int(np.ceil(image_shape[0] / roi_size[0]))
    n1 = int(np.ceil(image_shape[1] / roi_size[0]))
    for off0 in range(n0):
        for off1 in range(n1):

            start0 = pads[0][0] + off0 * roi_size[0]
            end0 = start0 + roi_size[0]
            start1 = pads[1][0] + off1 * roi_size[1]
            end1 = start1 + roi_size[1]
            if off0 == 0 and off1 > 0 and off1 < (n1 - 1):
                image[0:end0, start1:end1] = patches[
                    entry_index, 0:-padding, padding:-padding, 0
                ]
            elif off0 > 0 and off0 < (n0 - 1) and off1 == 0:
                image[start0:end0, 0:end1] = patches[
                    entry_index, padding:-padding, 0:-padding, 0
                ]
            elif off0 == 0 and off1 == 0:
                image[0:end0, 0:end1] = patches[entry_index, 0:-padding, 0:-padding, 0]
            elif off0 > 0 and off0 < (n0 - 1) and off1 == (n1 - 1):
                roi = patches[entry_index, padding:-padding, padding:, 0]
                image[start0:end0, start1 : (start1 + roi.shape[1])] = roi
            elif off0 == (n0 - 1) and off1 > 0 and off1 < (n1 - 1):
                roi = patches[entry_index, padding:, padding:-padding, 0]
                image[start0 : (start0 + roi.shape[0]), start1:end1] = roi
            elif off0 == (n0 - 1) and off1 == (n1 - 1):
                roi = patches[entry_index, padding:, padding:, 0]
                image[
                    start0 : (start0 + roi.shape[0]), start1 : (start1 + roi.shape[1])
                ] = roi
            elif off0 == 0 and off1 == (n1 - 1):
                roi = patches[entry_index, :-padding, padding:, 0]
                image[:end0, start1 : (start1 + roi.shape[1])] = roi
            elif off0 == (n0 - 1) and off1 == 0:
                roi = patches[entry_index, padding:, :-padding, 0]
                image[start0 : (start0 + roi.shape[0]), :end1] = roi
            else:
                image[start0:end0, start1:end1] = patches[
                    entry_index, padding:-padding, padding:-padding, 0
                ]
            entry_index = entry_index + 1
    image = image[pads[0][0] : -pads[0][1], pads[1][0] : -pads[1][1]]
    return image


def create_image_pair(movie_path):
    '''
    Calculates averages based on even / odd frames in a movie
    :param movie_path: Path to movie
    :return: even and odd average
    '''

    data = utils.read_image(movie_path)
    even = np.sum(data[::2], axis=0).astype(np.float32)
    odd = np.sum(data[1::2], axis=0).astype(np.float32)

    return even, odd


def normalize(img):
    '''
    Normalize a 2D image. Furthermore it will limit the values to -3 and 3 standard deviations.
    :param img: Image to normalize (2D numpy array)
    :return:  Normalized image, mean, standard diviation
    '''
    mean = np.mean(img)
    sd = np.std(img)
    img = (img - mean) / sd
    img = np.clip(img, -3, 3)
    return img, mean, sd


def read_image(path):
    if path.endswith((".tif", ".tiff")):
        try:
            img = tifffile.memmap(path,mode="r")
        except ValueError:
            img = tifffile.imread(path)
        return img
    elif path.endswith(("mrc", "mrcs")):
        with mrcfile.mmap(path, permissive=True) as mrc:
            return mrc.data
    else:
        print("Image format not supported. File: ", path)
        return None


def is_movie(path):
    '''
    Checks if file is movie or not
    :param path: Path to file
    :return: True if movie.
    '''
    if path.endswith((".tif", ".tiff")):
        tif = tifffile.TiffFile(path)
        return len(tif.pages) > 1
    elif path.endswith(("mrc", "mrcs")):
        with mrcfile.mmap(path, permissive=True) as mrc:
            return mrc.data.ndim > 2 and mrc.data.shape[0] > 1



"""
MIT License

Copyright (c) 2019 Max Planck Institute of Molecular Physiology

Author: Thorsten Wagner (thorsten.wagner@mpi-dortmund.mpg.de)
Author: Luca Lusnig (luca.lusnig@mpi-dortmund.mpg.de)
Author: Fabian Schoenfeld (fabian.schoenfeld@mpi-dortmund.mpg.de)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from keras.utils import Sequence
from random import shuffle
import numpy as np
import mrcfile
from . import utils


class patch_pair_batch_generator(Sequence):
    def __init__(
        self, pair_a_images, pair_b_images, patch_size, batch_size=4, augment=False
    ):
        self.pair_a_images = pair_a_images
        self.pair_b_images = pair_b_images
        self.patch_size = patch_size
        self.batch_size = batch_size
        self.augment = augment

    def __len__(self):
        length = int(np.ceil(len(self.pair_a_images) / self.batch_size))
        return length

    def __getitem__(self, idx):
        # Extracts a random crop from each image
        img_first = idx * self.batch_size
        img_last = img_first + self.batch_size
        if img_last > len(self.pair_a_images):
            img_last = len(self.pair_a_images)

        x = []
        y = []
        for img_index in range(img_first, img_last):

            pair_a_img_path = self.pair_a_images[img_index]
            pair_b_img_path = self.pair_b_images[img_index]

            data = utils.read_image(pair_a_img_path)

            start0 = np.random.randint(0, data.shape[0] - self.patch_size[0] + 1)
            end0 = start0 + self.patch_size[0]

            start1 = np.random.randint(0, data.shape[1] - self.patch_size[1] + 1)
            end1 = start1 + self.patch_size[1]
            pair_a = data[start0:end0, start1:end1].astype(np.float32)

            data = utils.read_image(pair_b_img_path)

            pair_b = data[start0:end0, start1:end1].astype(np.float32)

            if self.augment:
                flip_selection = np.random.randint(0, 4)
                flip_vertical = flip_selection == 1
                flip_horizontal = flip_selection == 2
                flip_both = flip_selection == 3

                if flip_vertical:
                    pair_a = np.flip(pair_a, 1)
                    pair_b = np.flip(pair_b, 1)
                if flip_horizontal:
                    pair_a = np.flip(pair_a, 0)
                    pair_b = np.flip(pair_b, 0)
                if flip_both:
                    pair_a = np.flip(np.flip(pair_a, 0), 1)
                    pair_b = np.flip(np.flip(pair_b, 0), 1)

                rand_rotation = np.random.randint(4)
                pair_a = np.rot90(pair_a, k=rand_rotation)
                pair_b = np.rot90(pair_b, k=rand_rotation)

                if np.random.rand() > 0.5:
                    help = pair_b
                    pair_b = pair_a
                    pair_a = help

            pair_a, _, _ = utils.normalize(pair_a)
            pair_b, _, _ = utils.normalize(pair_b)
            x.append(pair_a)
            y.append(pair_b)
        x = np.array(x)
        x = x[:, :, :, np.newaxis]
        y = np.array(y)
        y = y[:, :, :, np.newaxis]

        return x, y

    def on_epoch_end(self):
        index = list(range(len(self.pair_a_images)))
        shuffle(index)
        self.pair_a_images = [self.pair_a_images[i] for i in index]
        self.pair_b_images = [self.pair_b_images[i] for i in index]



from setuptools import setup
import os
import re
import codecs
import janni
# Create new package with python setup.py sdist

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='janni',
    version=find_version("janni", "__init__.py"),
    python_requires='>3.4.0',
    packages=['janni'],
    url='https://github.com/MPI-Dortmund/sphire-janni',
    license='MIT',
    author='Thorsten Wagner',
    setup_requires=["Cython"],
    extras_require={
        'gpu': ['tensorflow-gpu == 1.10.1'],
        'cpu': ['tensorflow == 1.10.1']
    },
    install_requires=[
        "mrcfile >=1.0.0,<= 1.1.2",
        "Keras == 2.2.5",
        "numpy == 1.14.5",
        "h5py >= 2.5.0",
        "Pillow >= 6.0.0",
        "Cython",
        "imagecodecs-lite==2019.2.22",
        "tifffile==2019.7.2",
        "GooeyDev >= 1.0.3.2",
        "wxPython == 4.0.4",
    ],
    author_email='thorsten.wagner@mpi-dortmund.mpg.de',
    description='noise 2 noise for cryo em data',
    entry_points={
        'console_scripts': [
            'janni_denoise.py = janni.jmain:_main_'
        ]},
)
